"""Contains API representation for the MySkoda REST API."""

import asyncio
import json
import logging
from collections.abc import Callable
from dataclasses import dataclass

from aiohttp import ClientResponseError, ClientSession

from myskoda.anonymize import (
    anonymize_air_conditioning,
    anonymize_charging,
    anonymize_driving_range,
    anonymize_garage,
    anonymize_health,
    anonymize_info,
    anonymize_maintenance,
    anonymize_positions,
    anonymize_status,
    anonymize_trip_statistics,
    anonymize_url,
    anonymize_user,
)
from myskoda.models.charging import ChargeMode
from myskoda.models.garage import Garage
from myskoda.models.position import Position, PositionType

from .auth.authorization import Authorization
from .const import BASE_URL_SKODA, REQUEST_TIMEOUT_IN_SECONDS
from .models.air_conditioning import AirConditioning
from .models.charging import Charging
from .models.driving_range import DrivingRange
from .models.health import Health
from .models.info import Info
from .models.maintenance import Maintenance
from .models.position import Positions
from .models.status import Status
from .models.trip_statistics import TripStatistics
from .models.user import User

_LOGGER = logging.getLogger(__name__)


@dataclass
class GetEndpointResult[T]:
    url: str
    raw: str
    result: T


class RestApi:
    """API hub class that can perform all calls to the MySkoda API."""

    session: ClientSession
    authorization: Authorization

    def __init__(self, session: ClientSession, authorization: Authorization) -> None:  # noqa: D107
        self.session = session
        self.authorization = authorization

    def process_json(
        self,
        data: str,
        anonymize: bool,
        anonymization_fn: Callable[[dict], dict],
    ) -> str:
        """Process the raw json returned by the API with some preprocessor logic."""
        if not anonymize:
            return data
        parsed = json.loads(data)
        anonymized = anonymization_fn(parsed)
        return json.dumps(anonymized)

    async def _make_request(self, url: str, method: str, json: dict | None = None) -> str:
        try:
            async with asyncio.timeout(REQUEST_TIMEOUT_IN_SECONDS):
                async with self.session.request(
                    method=method,
                    url=f"{BASE_URL_SKODA}/api{url}",
                    headers=await self._headers(),
                    json=json,
                ) as response:
                    await response.text()  # Ensure response is fully read
                    response.raise_for_status()
                    return await response.text()
        except TimeoutError:
            _LOGGER.exception("Timeout while sending %s request to %s", method, url)
            raise
        except ClientResponseError as err:
            _LOGGER.exception("Invalid status for %s request to %s: %d", method, url, err.status)
            raise

    async def _make_get_request[T](self, url: str) -> str:
        return await self._make_request(url=url, method="GET")

    async def _make_post_request(self, url: str, json: dict | None = None) -> str:
        return await self._make_request(url=url, method="POST", json=json)

    async def _make_put_request(self, url: str, json: dict | None = None) -> str:
        return await self._make_request(url=url, method="PUT", json=json)

    async def get_info(self, vin: str, anonymize: bool = False) -> GetEndpointResult[Info]:
        """Retrieve information related to basic information for the specified vehicle."""
        url = f"/v2/garage/vehicles/{vin}?connectivityGenerations=MOD1&connectivityGenerations=MOD2&connectivityGenerations=MOD3&connectivityGenerations=MOD4"  # noqa: E501
        raw = self.process_json(
            data=await self._make_get_request(url),
            anonymize=anonymize,
            anonymization_fn=anonymize_info,
        )
        result = self._deserialize(raw, Info.from_json)
        url = anonymize_url(url, vin) if anonymize else url
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def get_charging(self, vin: str, anonymize: bool = False) -> GetEndpointResult[Charging]:
        """Retrieve information related to charging for the specified vehicle."""
        url = f"/v1/charging/{vin}"
        raw = self.process_json(
            data=await self._make_get_request(url),
            anonymize=anonymize,
            anonymization_fn=anonymize_charging,
        )
        result = self._deserialize(raw, Charging.from_json)
        url = anonymize_url(url, vin) if anonymize else url
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def get_status(self, vin: str, anonymize: bool = False) -> GetEndpointResult[Status]:
        """Retrieve the current status for the specified vehicle."""
        url = f"/v2/vehicle-status/{vin}"
        raw = self.process_json(
            data=await self._make_get_request(url),
            anonymize=anonymize,
            anonymization_fn=anonymize_status,
        )
        result = self._deserialize(raw, Status.from_json)
        url = anonymize_url(url, vin) if anonymize else url
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def get_air_conditioning(
        self, vin: str, anonymize: bool = False
    ) -> GetEndpointResult[AirConditioning]:
        """Retrieve the current air conditioning status for the specified vehicle."""
        url = f"/v2/air-conditioning/{vin}"
        raw = self.process_json(
            data=await self._make_get_request(url),
            anonymize=anonymize,
            anonymization_fn=anonymize_air_conditioning,
        )
        result = self._deserialize(raw, AirConditioning.from_json)
        url = anonymize_url(url, vin) if anonymize else url
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def get_positions(
        self, vin: str, anonymize: bool = False
    ) -> GetEndpointResult[Positions]:
        """Retrieve the current position for the specified vehicle."""
        url = f"/v1/maps/positions?vin={vin}"
        raw = self.process_json(
            data=await self._make_get_request(url),
            anonymize=anonymize,
            anonymization_fn=anonymize_positions,
        )
        result = self._deserialize(raw, Positions.from_json)
        url = anonymize_url(url, vin) if anonymize else url
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def get_driving_range(
        self, vin: str, anonymize: bool = False
    ) -> GetEndpointResult[DrivingRange]:
        """Retrieve estimated driving range for combustion vehicles."""
        url = f"/v2/vehicle-status/{vin}/driving-range"
        raw = self.process_json(
            data=await self._make_get_request(url),
            anonymize=anonymize,
            anonymization_fn=anonymize_driving_range,
        )
        result = self._deserialize(raw, DrivingRange.from_json)
        url = anonymize_url(url, vin) if anonymize else url
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def get_trip_statistics(
        self, vin: str, anonymize: bool = False
    ) -> GetEndpointResult[TripStatistics]:
        """Retrieve statistics about past trips."""
        url = f"/v1/trip-statistics/{vin}?offsetType=week&offset=0&timezone=Europe%2FBerlin"
        raw = self.process_json(
            data=await self._make_get_request(url),
            anonymize=anonymize,
            anonymization_fn=anonymize_trip_statistics,
        )
        result = self._deserialize(raw, TripStatistics.from_json)
        url = anonymize_url(url, vin) if anonymize else url
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def get_maintenance(
        self, vin: str, anonymize: bool = False
    ) -> GetEndpointResult[Maintenance]:
        """Retrieve maintenance report."""
        url = f"/v3/vehicle-maintenance/vehicles/{vin}"
        raw = self.process_json(
            data=await self._make_get_request(url),
            anonymize=anonymize,
            anonymization_fn=anonymize_maintenance,
        )
        result = self._deserialize(raw, Maintenance.from_json)
        url = anonymize_url(url, vin) if anonymize else url
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def get_health(self, vin: str, anonymize: bool = False) -> GetEndpointResult[Health]:
        """Retrieve health information for the specified vehicle."""
        url = f"/v1/vehicle-health-report/warning-lights/{vin}"
        raw = self.process_json(
            data=await self._make_get_request(url),
            anonymize=anonymize,
            anonymization_fn=anonymize_health,
        )
        result = self._deserialize(raw, Health.from_json)
        url = anonymize_url(url, vin) if anonymize else url
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def get_user(self, anonymize: bool = False) -> GetEndpointResult[User]:
        """Retrieve user information about logged in user."""
        url = "/v1/users"
        raw = self.process_json(
            data=await self._make_get_request(url),
            anonymize=anonymize,
            anonymization_fn=anonymize_user,
        )
        result = self._deserialize(raw, User.from_json)
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def get_garage(self, anonymize: bool = False) -> GetEndpointResult[Garage]:
        """Fetch the garage (list of vehicles with limited info)."""
        url = "/v2/garage?connectivityGenerations=MOD1&connectivityGenerations=MOD2&connectivityGenerations=MOD3&connectivityGenerations=MOD4"  # noqa: E501
        raw = self.process_json(
            data=await self._make_get_request(url),
            anonymize=anonymize,
            anonymization_fn=anonymize_garage,
        )
        result = self._deserialize(raw, Garage.from_json)
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def _headers(self) -> dict[str, str]:
        return {"authorization": f"Bearer {await self.authorization.get_access_token()}"}

    async def stop_air_conditioning(self, vin: str) -> None:
        """Stop the air conditioning."""
        _LOGGER.debug("Stopping air conditioning for vehicle %s", vin)
        await self._make_post_request(url=f"/v2/air-conditioning/{vin}/stop")

    async def start_air_conditioning(self, vin: str, temperature: float) -> None:
        """Start the air conditioning."""
        _LOGGER.debug(
            "Starting air conditioning for vehicle %s with temperature %s",
            vin,
            str(temperature),
        )
        json_data = {
            "heaterSource": "ELECTRIC",
            "targetTemperature": {
                "temperatureValue": str(temperature),
                "unitInCar": "CELSIUS",
            },
        }
        await self._make_post_request(
            url=f"/v2/air-conditioning/{vin}/start",
            json=json_data,
        )

    async def set_target_temperature(self, vin: str, temperature: float) -> None:
        """Set the air conditioning's target temperature in °C."""
        _LOGGER.debug("Setting target temperature for vehicle %s to %s", vin, str(temperature))
        json_data = {"temperatureValue": str(temperature), "unitInCar": "CELSIUS"}
        await self._make_post_request(
            url=f"/v2/air-conditioning/{vin}/settings/target-temperature",
            json=json_data,
        )

    async def start_window_heating(self, vin: str) -> None:
        """Start heating both the front and rear window."""
        _LOGGER.debug("Starting window heating for vehicle %s", vin)
        await self._make_post_request(
            url=f"/v2/air-conditioning/{vin}/start-window-heating",
        )

    async def stop_window_heating(self, vin: str) -> None:
        """Stop heating both the front and rear window."""
        _LOGGER.debug("Stopping window heating for vehicle %s", vin)
        await self._make_post_request(
            url=f"/v2/air-conditioning/{vin}/stop-window-heating",
        )

    async def set_charge_limit(self, vin: str, limit: int) -> None:
        """Set the maximum charge limit in percent."""
        _LOGGER.debug("Setting charge limit for vehicle %s to %d", vin, limit)
        json_data = {"targetSOCInPercent": limit}
        await self._make_put_request(
            url=f"/v1/charging/{vin}/set-charge-limit",
            json=json_data,
        )

    # TODO @dvx76: Maybe refactor for FBT001
    async def set_battery_care_mode(self, vin: str, enabled: bool) -> None:
        """Enable or disable the battery care mode."""
        _LOGGER.debug("Setting battery care mode for vehicle %s to %r", vin, enabled)
        json_data = {"chargingCareMode": "ACTIVATED" if enabled else "DEACTIVATED"}
        await self._make_put_request(
            url=f"/v1/charging/{vin}/set-care-mode",
            json=json_data,
        )

    # TODO @dvx76: Maybe refactor for FBT001
    async def set_reduced_current_limit(self, vin: str, reduced: bool) -> None:
        """Enable reducing the current limit by which the car is charged."""
        _LOGGER.debug("Setting reduced charging for vehicle %s to %r", vin, reduced)
        json_data = {"chargingCurrent": "REDUCED" if reduced else "MAXIMUM"}
        await self._make_put_request(
            url=f"/v1/charging/{vin}/set-charging-current",
            json=json_data,
        )

    async def start_charging(self, vin: str) -> None:
        """Start charging the car."""
        _LOGGER.debug("Starting charging for vehicle %s", vin)
        await self._make_post_request(
            url=f"/v1/charging/{vin}/start",
        )

    async def stop_charging(self, vin: str) -> None:
        """Stop charging the car."""
        _LOGGER.debug("Stopping charging of vehicle %s", vin)
        await self._make_post_request(
            url=f"/v1/charging/{vin}/stop",
        )

    async def wakeup(self, vin: str) -> None:
        """Wake the vehicle up. Can be called maximum three times a day."""
        _LOGGER.debug("Waking up vehicle %s", vin)
        await self._make_post_request(
            url=f"/v1/vehicle-wakeup/{vin}?applyRequestLimiter=true",
        )

    async def set_charge_mode(self, vin: str, mode: ChargeMode) -> None:
        """Wake the vehicle up. Can be called maximum three times a day."""
        _LOGGER.debug("Changing charging mode of vehicle %s to %s", vin, mode)
        json_data = {"chargeMode": mode.value}
        await self._make_post_request(
            url=f"/v1/charging/{vin}/set-charge-mode",
            json=json_data,
        )

    # TODO @dvx76: Maybe refactor for FBT001
    async def honk_flash(
        self,
        vin: str,
        positions: list[Position],
        honk: bool = False,
    ) -> None:
        """Honk and/or flash."""
        position = next(pos for pos in positions if pos.type == PositionType.VEHICLE)
        json_data = {
            "mode": "HONK_AND_FLASH" if honk else "FLASH",
            "vehiclePosition": {
                "lat": position.gps_coordinates.latitude,
                "lng": position.gps_coordinates.longitude,
            },
        }
        await self._make_post_request(
            url=f"/v1/vehicle-access/{vin}/honk-and-flash", json=json_data
        )

    def _deserialize[T](self, text: str, deserialize: Callable[[str], T]) -> T:
        try:
            data = deserialize(text)
        except Exception:
            _LOGGER.exception("Failed to deserialize data: %s", text)
            raise
        else:
            return data



"""CLI to test all API functions and models.

Execute with:
poetry run python3 -m myskoda.cli
"""

from logging import DEBUG, INFO
from sys import platform as sys_platform
from sys import version_info as sys_version_info

import asyncclick as click
import coloredlogs
from aiohttp import ClientSession
from asyncclick.core import Context

from myskoda import TRACE_CONFIG, MySkoda
from myskoda.cli.gen_fixtures import gen_fixtures
from myskoda.cli.mqtt import subscribe, wait_for_operation
from myskoda.cli.operations import (
    set_charge_limit,
    set_reduced_current_limit,
    set_target_temperature,
    start_air_conditioning,
    start_window_heating,
    stop_air_conditioning,
    stop_window_heating,
    wakeup,
)
from myskoda.cli.requests import (
    air_conditioning,
    auth,
    charging,
    driving_range,
    garage,
    health,
    info,
    list_vehicles,
    maintenance,
    positions,
    status,
    trip_statistics,
    user,
)
from myskoda.cli.utils import Format, print_json, print_yaml


@click.group()
@click.version_option()
@click.option("username", "--user", help="Username used for login.", required=True)
@click.option("password", "--password", help="Password used for login.", required=True)
@click.option("verbose", "--verbose", help="Enable verbose logging.", is_flag=True)
@click.option(
    "output_format",
    "--format",
    help="Select the output format. JSON or YAML.",
    type=click.Choice(Format),  # pyright: ignore [reportArgumentType]
    default=Format.YAML,
)
@click.option("trace", "--trace", help="Enable tracing of HTTP requests.", is_flag=True)
@click.option("disable_mqtt", "--disable-mqtt", help="Do not connect to MQTT.", is_flag=True)
@click.pass_context
async def cli(  # noqa: PLR0913
    ctx: Context,
    username: str,
    password: str,
    verbose: bool,
    output_format: Format,
    trace: bool,
    disable_mqtt: bool,
) -> None:
    """Interact with the MySkoda API."""
    coloredlogs.install(level=DEBUG if verbose else INFO)
    ctx.ensure_object(dict)
    ctx.obj["username"] = username
    ctx.obj["password"] = password
    if output_format == Format.JSON:
        ctx.obj["print"] = print_json
    elif output_format == Format.YAML:
        ctx.obj["print"] = print_yaml

    ctx.obj["mqtt_disabled"] = disable_mqtt

    trace_configs = []
    if trace:
        trace_configs.append(TRACE_CONFIG)

    # Check if we're on windows, if so, tune asyncio to work there as well (https://github.com/skodaconnect/myskoda/issues/77)
    if sys_platform.lower().startswith("win") and sys_version_info >= (3, 8):
        import asyncio

        try:
            from asyncio import WindowsSelectorEventLoopPolicy  # type: ignore[unknown-import]
        except ImportError:
            pass  # Can't assign a policy which doesn't exist.
        else:
            if not isinstance(asyncio.get_event_loop_policy(), WindowsSelectorEventLoopPolicy):
                asyncio.set_event_loop_policy(WindowsSelectorEventLoopPolicy())

    session = ClientSession(trace_configs=trace_configs)
    myskoda = MySkoda(session, mqtt_enabled=False)
    await myskoda.connect(username, password)

    ctx.obj["myskoda"] = myskoda
    ctx.obj["session"] = session


@cli.result_callback()
@click.pass_context
async def disconnect(  # noqa: PLR0913
    ctx: Context,
    result: None,  # noqa: ARG001
    username: str,  # noqa: ARG001
    password: str,  # noqa: ARG001
    verbose: bool,  # noqa: ARG001
    output_format: Format,  # noqa: ARG001
    trace: bool,  # noqa: ARG001
    disable_mqtt: bool,  # noqa: ARG001
) -> None:
    myskoda: MySkoda = ctx.obj["myskoda"]
    session: ClientSession = ctx.obj["session"]

    myskoda.disconnect()
    await session.close()


cli.add_command(list_vehicles)
cli.add_command(info)
cli.add_command(status)
cli.add_command(air_conditioning)
cli.add_command(positions)
cli.add_command(health)
cli.add_command(charging)
cli.add_command(maintenance)
cli.add_command(driving_range)
cli.add_command(user)
cli.add_command(trip_statistics)
cli.add_command(garage)
cli.add_command(auth)
cli.add_command(start_air_conditioning)
cli.add_command(stop_air_conditioning)
cli.add_command(set_target_temperature)
cli.add_command(start_window_heating)
cli.add_command(stop_window_heating)
cli.add_command(set_charge_limit)
cli.add_command(set_reduced_current_limit)
cli.add_command(wakeup)
cli.add_command(wait_for_operation)
cli.add_command(subscribe)
cli.add_command(gen_fixtures)

if __name__ == "__main__":
    cli()



"""User that is using the API."""

from dataclasses import dataclass, field
from datetime import date
from enum import StrEnum

from mashumaro import field_options
from mashumaro.mixins.orjson import DataClassORJSONMixin


class UserCapabilityId(StrEnum):
    SPIN_MANAGEMENT = "SPIN_MANAGEMENT"
    THIRD_PARTY_OFFERS = "THIRD_PARTY_OFFERS"
    MARKETING_CONSENT = "MARKETING_CONSENT"
    TEST_DRIVE = "TEST_DRIVE"


@dataclass
class UserCapability(DataClassORJSONMixin):
    id: UserCapabilityId


@dataclass
class User(DataClassORJSONMixin):
    capabilities: list[UserCapability]
    country: str
    email: str
    first_name: str = field(metadata=field_options(alias="firstName"))
    id: str
    last_name: str = field(metadata=field_options(alias="lastName"))
    nickname: str
    preferred_language: str = field(metadata=field_options(alias="preferredLanguage"))
    profile_picture_url: str = field(metadata=field_options(alias="profilePictureUrl"))
    date_of_birth: date | None = field(default=None, metadata=field_options(alias="dateOfBirth"))
    preferred_contact_channel: str | None = field(
        default=None, metadata=field_options(alias="preferredContactChannel")
    )
    phone: str | None = None



"""Models for responses of api/v2/garage/vehicles/{vin}."""

import logging
from dataclasses import dataclass, field

from mashumaro import field_options
from mashumaro.mixins.orjson import DataClassORJSONMixin

from myskoda.models.info import CompositeRender, Render, VehicleState

_LOGGER = logging.getLogger(__name__)


@dataclass
class GarageEntry(DataClassORJSONMixin):
    """One vehicle in the list of vehicles."""

    vin: str
    name: str
    state: VehicleState
    title: str
    priority: int
    device_platform: str = field(metadata=field_options(alias="devicePlatform"))
    system_model_id: str = field(metadata=field_options(alias="systemModelId"))
    renders: list[Render]
    composite_renders: list[CompositeRender] = field(
        metadata=field_options(alias="compositeRenders")
    )


@dataclass
class Garage(DataClassORJSONMixin):
    """Contents of the users Garage."""

    vehicles: list[GarageEntry] | None = field(default=None)



"""Generate a set of test fixtures from your garage."""

from pathlib import Path
from typing import cast

import asyncclick as click
from asyncclick.core import Context

from myskoda.models.fixtures import Endpoint
from myskoda.myskoda import MySkoda

FIXTURES_DIR = Path(__file__).parent.parent.parent / "tests" / "fixtures" / "gen"


@click.group()
@click.option(
    "vehicle",
    "--vehicle",
    help="Generate fixtures for this vehicle only (vin). Set to 'all' for all vehicles.",
    required=True,
)
@click.option("name", "--name", help="Short name describing the vehicle's state.", required=True)
@click.option("description", "--description", help="A longer description.")
@click.option("file", "--file", help="Override default file name.")
@click.pass_context
async def gen_fixtures(
    ctx: Context, vehicle: str, name: str, description: str, file: str | None
) -> None:
    """Interact with the MySkoda API."""
    myskoda: MySkoda = ctx.obj["myskoda"]

    default_filename = Path("fixtures") / f"{name.replace(" ", "_").lower()}.yaml"

    ctx.obj["vins"] = await get_vin_list(myskoda, vehicle)
    ctx.obj["name"] = name
    ctx.obj["description"] = description
    ctx.obj["file"] = Path(file) if file is not None else default_filename


@gen_fixtures.command()
@click.argument("endpoint", type=click.Choice(Endpoint))  # pyright: ignore [reportArgumentType]
@click.pass_context
async def get(ctx: Context, endpoint: Endpoint) -> None:
    myskoda: MySkoda = ctx.obj["myskoda"]

    fixture = await myskoda.generate_get_fixture(
        name=ctx.obj["name"],
        description=ctx.obj["description"],
        vins=ctx.obj["vins"],
        endpoint=endpoint,
    )

    text = cast(str, fixture.to_yaml())

    file: Path = ctx.obj["file"]
    file.parent.mkdir(parents=True, exist_ok=True)
    file.write_text(text)


async def get_vin_list(myskoda: MySkoda, vehicle: str) -> list[str]:
    if vehicle != "all":
        return [vehicle]
    return await myskoda.list_vehicle_vins()



"""Handles authorization to the MySkoda API."""

import base64
import hashlib
import logging
from asyncio import Lock
from collections.abc import Callable
from dataclasses import dataclass, field
from datetime import UTC, datetime, timedelta
from typing import cast

import jwt
from aiohttp import ClientSession, FormData, InvalidUrlClientError
from mashumaro import field_options
from mashumaro.mixins.orjson import DataClassORJSONMixin

from myskoda.auth.csrf_parser import CSRFParser, CSRFState
from myskoda.auth.utils import generate_nonce
from myskoda.const import BASE_URL_IDENT, BASE_URL_SKODA, CLIENT_ID, MAX_RETRIES

_LOGGER = logging.getLogger(__name__)


@dataclass
class IDKAuthorizationCode(DataClassORJSONMixin):
    """One-time authorization code that can be obtained by logging in.

    This authorization code can later be exchanged for a set of JWT tokens.
    """

    code: str
    token_type: str
    id_token: str


refresh_token_lock = Lock()


@dataclass
class IDKSession(DataClassORJSONMixin):
    """Stores the JWT tokens relevant for a session at the IDK server.

    Can be used to authorized and refresh the authorization token.
    """

    access_token: str = field(metadata=field_options(alias="accessToken"))
    refresh_token: str = field(metadata=field_options(alias="refreshToken"))
    id_token: str = field(metadata=field_options(alias="idToken"))


class Authorization:
    """Class that holds Authorization information and authorization state of the session."""

    session: ClientSession
    idk_session: IDKSession | None = None

    def __init__(  # noqa: D107
        self, session: ClientSession, generate_nonce: Callable[[], str] = generate_nonce
    ) -> None:
        self.session = session
        self.generate_nonce = generate_nonce

    def _extract_csrf(self, html: str) -> CSRFState:
        parser = CSRFParser()
        parser.feed(html)

        if parser.csrf_state is None:
            raise CSRFError

        return parser.csrf_state

    async def authorize(self, email: str, password: str) -> None:
        """Authorize on the VW IDK servers."""
        self.email = email
        self.password = password
        try:
            self.idk_session = await self._get_idk_session()
        except InvalidUrlClientError as ex:
            raise AuthorizationFailedError from ex

        if self.idk_session is None:
            raise AuthorizationFailedError

    async def _initial_oidc_authorize(self, verifier: str) -> CSRFState:
        """First step of the login process.

        This calls the route for initial authorization,
        which will contain the initial SSO information such as the CSRF or the HMAC.
        """
        # A SHA256 hash of the random "verifier" string will be transmitted as a challenge.
        # This is part of the OAUTH2 PKCE process. It is described here in detail:
        # https://www.oauth.com/oauth2-servers/pkce/authorization-request/
        verifier_hash = hashlib.sha256(verifier.encode("utf-8")).digest()
        challenge = (
            base64.b64encode(verifier_hash)
            .decode("utf-8")
            .replace("+", "-")
            .replace("/", "_")
            .rstrip("=")
        )

        params = {
            "client_id": CLIENT_ID,
            "nonce": self.generate_nonce(),
            "redirect_uri": "myskoda://redirect/login/",
            "response_type": "code id_token",
            # OpenID scopes. Can be found here: https://identity.vwgroup.io/.well-known/openid-configuration
            "scope": "address badge birthdate cars driversLicense dealers email mileage mbb nationalIdentifier openid phone profession profile vin",  # noqa: E501
            "code_challenge": challenge,
            "code_challenge_method": "s256",
            "prompt": "login",
        }
        async with self.session.get(
            f"{BASE_URL_IDENT}/oidc/v1/authorize", params=params
        ) as response:
            return self._extract_csrf(await response.text())

    async def _enter_email_address(self, csrf: CSRFState) -> CSRFState:
        """Second step in the login process.

        Will post only the email address to the backend.
        The password will follow in a later request.
        """
        form_data = FormData()
        form_data.add_field("relayState", csrf.template_model.relay_state)
        form_data.add_field("email", self.email)
        form_data.add_field("hmac", csrf.template_model.hmac)
        form_data.add_field("_csrf", csrf.csrf)

        async with self.session.post(
            f"{BASE_URL_IDENT}/signin-service/v1/{CLIENT_ID}/login/identifier",
            data=form_data(),
        ) as response:
            return self._extract_csrf(await response.text())

    async def _enter_password(self, csrf: CSRFState) -> IDKAuthorizationCode:
        """Third step in the login process.

        Post both the email address and the password to the backend.
        This will return a token which can then be used in the skoda services to authenticate.
        """
        form_data = FormData()
        form_data.add_field("relayState", csrf.template_model.relay_state)
        form_data.add_field("email", self.email)
        form_data.add_field("password", self.password)
        form_data.add_field("hmac", csrf.template_model.hmac)
        form_data.add_field("_csrf", csrf.csrf)

        # The following is a bit hacky:
        # The backend will redirect multiple times after the login was successful.
        # The last redirect will redirect back to the `MySkoda` app in Android,
        # using the `myskoda://` URL prefix.
        # The following loop will follow all redirects until the last redirect to `myskoda://` is
        # encountered. This last URL will contain the token.
        try:
            async with self.session.post(
                f"{BASE_URL_IDENT}/signin-service/v1/{CLIENT_ID}/login/authenticate",
                data=form_data(),
                allow_redirects=False,
                raise_for_status=True,
            ) as auth_response:
                location = auth_response.headers["Location"]
                while not location.startswith("myskoda://"):
                    async with self.session.get(location, allow_redirects=False) as response:
                        location = response.headers["Location"]
                codes = location.replace("myskoda://redirect/login/#", "")
        except InvalidUrlClientError:
            _LOGGER.exception("Error occurred while sending password. Password may be incorrect.")
            raise

        # The last redirection starting with `myskoda://` was encountered.
        # The URL will contain the information we need as query parameters,
        # without the leading `?`.
        data = {}
        for code in codes.split("&"):
            [key, value] = code.split("=")
            data[key] = value

        return IDKAuthorizationCode.from_dict(data)

    async def _exchange_auth_code_for_idk_session(self, code: str, verifier: str) -> IDKSession:
        """Exchange the ident login code for an auth token from Skoda.

        This will return multiple tokens, such as an access token and a refresh token.
        """
        json_data = {
            "code": code,
            "redirectUri": "myskoda://redirect/login/",
            "verifier": verifier,
        }

        async with self.session.post(
            f"{BASE_URL_SKODA}/api/v1/authentication/exchange-authorization-code?tokenType=CONNECT",
            json=json_data,
            allow_redirects=False,
        ) as response:
            return IDKSession.from_json(await response.text())

    async def _get_idk_session(self) -> IDKSession:
        """Perform the full login process.

        Must be called before any other methods on the class can be called.
        """
        # Generate a random string for the OAUTH2 PKCE challenge.
        # (https://www.oauth.com/oauth2-servers/pkce/authorization-request/)
        verifier = self.generate_nonce()

        # Call the initial OIDC (OpenID Connect) authorization,
        # giving us the initial SSO information.
        # The full flow is explain a little bit here:
        # https://openid.net/specs/openid-connect-core-1_0.html#ImplicitFlowAuth
        login_meta = await self._initial_oidc_authorize(verifier)

        # Use the information to login with the email address,
        # which is an extra step before the actual login.
        login_meta = await self._enter_email_address(login_meta)

        # Perform the actual login which will result in a token that can be exchanged for
        # an access token at the Skoda server.
        authentication = await self._enter_password(login_meta)

        # Exchange the token for access and refresh tokens (JWT format).
        return await self._exchange_auth_code_for_idk_session(authentication.code, verifier)

    def is_token_expired(self) -> bool:
        """Check whether the login token is expired."""
        if not self.idk_session:
            raise NotAuthorizedError

        meta = jwt.decode(self.idk_session.access_token, options={"verify_signature": False})
        expiry = datetime.fromtimestamp(cast(float, meta.get("exp")), tz=UTC)
        return datetime.now(tz=UTC) + timedelta(minutes=10) > expiry

    async def _perform_refresh_token(self) -> bool:
        if not self.idk_session:
            raise NotAuthorizedError

        if not self.is_token_expired():
            return True

        async with self.session.post(
            f"{BASE_URL_SKODA}/api/v1/authentication/refresh-token?tokenType=CONNECT",
            json={"token": self.idk_session.refresh_token},
        ) as response:
            if not response.ok:
                return False
            try:
                self.idk_session = IDKSession.from_json(await response.text())
            except Exception:
                _LOGGER.exception("Failed to parse tokens from refresh endpoint.")
                return False
            else:
                return True

    async def refresh_token(self) -> None:
        """Refresh the authorization token.

        This will consume the `refresh_token` and exchange it for a new set of tokens.
        """
        async with refresh_token_lock:
            for attempt in range(MAX_RETRIES):
                if await self._perform_refresh_token():
                    return
                _LOGGER.warning(
                    "Retrying failed request to refresh token (%d/%d). Retrying...",
                    attempt,
                    MAX_RETRIES,
                )

            _LOGGER.error("Refreshing token failed after %d attempts.", MAX_RETRIES)
            _LOGGER.info("Trying to recover by logging in again...")

            try:
                idk_session = await self._get_idk_session()
            except Exception:
                _LOGGER.exception("Failed to login.")
            else:
                self.idk_session = idk_session
                _LOGGER.info("Successfully recovered by logging in.")
                return

    async def get_access_token(self) -> str:
        """Get the access token.

        Use this method instead of using `access_token` directly. It will automatically
        check if the JWT token is about to expire and refresh it using the `refresh_token`.
        """
        if self.idk_session is None:
            raise NotAuthorizedError

        if self.is_token_expired():
            _LOGGER.info("Token expired. Refreshing IDK access token")
            await self.refresh_token()
        return self.idk_session.access_token


class AuthorizationError(Exception):
    """Error to indicate that something unexpected happened during authorization."""


class InvalidStatusError(Exception):
    """An invalid HTTP status code was received."""

    def __init__(self, status: int) -> None:  # noqa: D107
        super().__init__(f"Received invalid HTTP status code {status}.")


class CSRFError(Exception):
    """Failed to parse the CSRF information from the website."""


class NotAuthorizedError(Exception):
    """Not authorized.

    Did you forget to call Authorization.authorize()?
    """


class AuthorizationFailedError(Exception):
    """Failed to authorize."""



"""MQTT client module for the MySkoda server."""

import json
import logging
import re
import ssl
from asyncio import Future, Lock, create_task, get_event_loop, sleep
from collections.abc import Awaitable, Callable
from typing import Any, cast

from asyncio_paho.client import AsyncioPahoClient
from paho.mqtt.client import MQTTMessage

from myskoda.auth.authorization import Authorization

from .const import (
    MQTT_ACCOUNT_EVENT_TOPICS,
    MQTT_BROKER_HOST,
    MQTT_BROKER_PORT,
    MQTT_KEEPALIVE,
    MQTT_OPERATION_TOPICS,
    MQTT_RECONNECT_DELAY,
    MQTT_SERVICE_EVENT_TOPICS,
)
from .event import (
    Event,
    EventAccess,
    EventAccountPrivacy,
    EventAirConditioning,
    EventCharging,
    EventLights,
    EventOperation,
    EventType,
)
from .models.operation_request import OperationName, OperationRequest, OperationStatus

_LOGGER = logging.getLogger(__name__)
TOPIC_RE = re.compile("^(.*?)/(.*?)/(.*?)/(.*?)$")

background_tasks = set()


class OperationListener:
    operation_name: OperationName
    future: Future[OperationRequest]

    def __init__(  # noqa: D107
        self, operation_name: OperationName, future: Future[OperationRequest]
    ) -> None:
        self.operation_name = operation_name
        self.future = future


connect_lock = Lock()


class Mqtt:
    user_id: str | None
    vehicles: list[str] | None
    client: AsyncioPahoClient
    _callbacks: list[Callable[[Event], None | Awaitable[None]]]
    _operation_listeners: list[OperationListener]
    _connected_listeners: list[Future[None]]
    should_reconnect: bool
    is_connected: bool
    authorization: Authorization

    def __init__(  # noqa: D107
        self,
        authorization: Authorization,
        ssl_context: ssl.SSLContext | None = None,
    ) -> None:
        self.authorization = authorization
        self._callbacks = []
        self._operation_listeners = []
        self._connected_listeners = []
        self.ssl_context = ssl_context
        self.is_connected = False
        self.user_id = None
        self.vehicles = None

    def subscribe(self, callback: Callable[[Event], None | Awaitable[None]]) -> None:
        """Listen for events emitted by MySkoda's MQTT broker."""
        self._callbacks.append(callback)

    async def _perform_connect(self) -> bool:
        """Connect to the MQTT broker once.

        Will return `True` if the connection was established and `False` otherwise.
        """
        try:
            if self.is_connected:
                return True

            _LOGGER.debug("Connecting to MQTT on %s:%d...", MQTT_BROKER_HOST, MQTT_BROKER_PORT)

            self.should_reconnect = True

            self.client = AsyncioPahoClient()
            self.client.on_connect = self._on_connect
            self.client.on_message = self._on_message
            self.client.on_disconnect = self._on_disconnect
            self.client.on_socket_close = self._on_socket_close
            self.client.on_connect_fail = self._on_connect_fail
            if self.ssl_context is not None:
                self.client.tls_set_context(context=self.ssl_context)
            else:
                self.client.tls_set_context(context=ssl.create_default_context())

            self.client.username_pw_set(
                self.user_id,
                await self.authorization.get_access_token(),
            )
            self.client.connect_async(MQTT_BROKER_HOST, MQTT_BROKER_PORT, MQTT_KEEPALIVE)

            await self._wait_for_connection()
        except FailedToConnectError:
            return False
        else:
            return True

    async def connect(self, user_id: str, vehicles: list[str]) -> None:
        """Connect to the MQTT broker and listen for messages."""
        self.user_id = user_id
        self.vehicles = vehicles
        async with connect_lock:
            while not await self._perform_connect():  # noqa: ASYNC110
                await sleep(MQTT_RECONNECT_DELAY)

    def reconnect(self) -> None:
        """Reconnect a client that was previously connected and was disconnected."""
        _LOGGER.debug("Scheduling to reconnect MQTT.")

        if not self.should_reconnect:
            return

        if not self.user_id or not self.vehicles:
            raise NotConnectedError

        task = create_task(cast(Any, self.connect(self.user_id, self.vehicles)))
        background_tasks.add(task)
        task.add_done_callback(background_tasks.discard)

    def disconnect(self) -> None:
        """Stop the thread for processing MQTT messages."""
        self.should_reconnect = False
        self.client.disconnect()  # pyright: ignore [reportArgumentType]

    def _wait_for_connection(self) -> Future[None]:
        """Wait until MQTT is connected and setup."""
        future: Future[None] = get_event_loop().create_future()

        self._connected_listeners.append(future)

        return future

    def wait_for_operation(self, operation_name: OperationName) -> Future[OperationRequest]:
        """Wait until the next operation of the specified type completes."""
        _LOGGER.debug("Waiting for operation %s complete.", operation_name)
        future: Future[OperationRequest] = get_event_loop().create_future()

        self._operation_listeners.append(OperationListener(operation_name, future))

        return future

    def _on_socket_close(self, client: AsyncioPahoClient, _data: None, _socket: None) -> None:
        if client is not self.client:
            return
        _LOGGER.debug("Socket to MQTT broker closed.")
        self.is_connected = False
        self.reconnect()

    def _on_connect_fail(self, client: AsyncioPahoClient, _data: None) -> None:
        if client is not self.client:
            return
        _LOGGER.error("Failed to connect to MQTT.")
        for future in self._connected_listeners:
            future.set_exception(FailedToConnectError)
        self._connected_listeners = []

    def _on_disconnect(
        self,
        client: AsyncioPahoClient,
        _userdata: None,
        reason_code: int,
    ) -> None:
        if client is not self.client:
            return
        _LOGGER.debug("Connection to MQTT broker lost, reason %d.", reason_code)
        self.is_connected = False
        self.reconnect()

    def _on_connect(
        self, client: AsyncioPahoClient, _data: None, _flags: dict, _reason: int
    ) -> None:
        if client is not self.client:
            return

        self.is_connected = True

        _LOGGER.debug("MQTT Connected.")
        if not self.user_id or not self.vehicles:
            _LOGGER.error("Reached on_connect, but user and vehicles not loaded")
            return

        user_id = self.user_id

        for vin in self.vehicles:
            for topic in MQTT_OPERATION_TOPICS:
                self._subscribe_to_topic(f"{user_id}/{vin}/operation-request/{topic}")
            for topic in MQTT_SERVICE_EVENT_TOPICS:
                self._subscribe_to_topic(f"{user_id}/{vin}/service-event/{topic}")
            for topic in MQTT_ACCOUNT_EVENT_TOPICS:
                self._subscribe_to_topic(f"{user_id}/{vin}/account-event/{topic}")

        for future in self._connected_listeners:
            future.set_result(None)
        self._connected_listeners = []

    def _subscribe_to_topic(self, topic: str) -> None:
        _LOGGER.debug("Subscribing to topic: %s", topic)
        self.client.subscribe(topic)

    def _emit(self, event: Event) -> None:
        for callback in self._callbacks:
            result = callback(event)
            if result is not None:
                task = create_task(cast(Any, result))
                background_tasks.add(task)
                task.add_done_callback(background_tasks.discard)

        self._handle_operation(event)

    def _handle_operation_completed(self, operation: OperationRequest) -> None:
        listeners = self._operation_listeners
        self._operation_listeners = []
        for listener in listeners:
            if listener.operation_name != operation.operation:
                self._operation_listeners.append(listener)
                continue

            if operation.status == OperationStatus.ERROR:
                _LOGGER.error(
                    "Resolving listener for operation '%s' with error '%s'.",
                    operation.operation,
                    operation.error_code,
                )
                listener.future.set_exception(OperationFailedError(operation))
            else:
                if operation.status == OperationStatus.COMPLETED_WARNING:
                    _LOGGER.warning("Operation '%s' completed with warnings.", operation.operation)

                _LOGGER.debug("Resolving listener for operation '%s'.", operation.operation)
                listener.future.set_result(operation)

    def _handle_operation(self, event: Event) -> None:
        if event.type != EventType.OPERATION:
            return

        if event.operation.status == OperationStatus.IN_PROGRESS:
            _LOGGER.debug(
                "An operation '%s' is now in progress. Trace id: %s",
                event.operation.operation,
                event.operation.trace_id,
            )
            return

        _LOGGER.debug(
            "Operation '%s' for trace id '%s' completed.",
            event.operation.operation,
            event.operation.trace_id,
        )
        self._handle_operation_completed(event.operation)

    def _on_message(  # noqa: C901
        self, client: AsyncioPahoClient, _data: None, msg: MQTTMessage
    ) -> None:
        if client is not self.client:
            return

        # Extract the topic, user id and vin from the topic's name.
        # Internally, the topic will always look like this:
        # `/{user_id}/{vin}/path/to/topic`
        topic_match = TOPIC_RE.match(msg.topic)
        if not topic_match:
            _LOGGER.warning("Unexpected MQTT topic encountered: %s", topic_match)
            return

        [user_id, vin, event_type, topic] = topic_match.groups()
        event_type = EventType(event_type)

        # Cast the data from binary string, ignoring empty messages.
        data = cast(str, msg.payload)
        if len(data) == 0:
            return

        _LOGGER.debug("Message (%s) received for %s on topic %s: %s", event_type, vin, topic, data)

        # Messages will contain payload as JSON.
        try:
            data = json.loads(msg.payload)
            match event_type:
                case EventType.OPERATION:
                    self._emit(EventOperation(vin, user_id, data))
                case EventType.ACCOUNT_EVENT:
                    self._emit(EventAccountPrivacy(vin, user_id, data))
                case EventType.SERVICE_EVENT:
                    match topic:
                        case "air-conditioning":
                            self._emit(EventAirConditioning(vin, user_id, data))
                        case "charging":
                            self._emit(EventCharging(vin, user_id, data))
                        case "vehicle-status/access":
                            self._emit(EventAccess(vin, user_id, data))
                        case "vehicle-status/lights":
                            self._emit(EventLights(vin, user_id, data))
        except Exception as exc:  # noqa: BLE001
            _LOGGER.warning("Exception parsing MQTT event: %s", exc)


class OperationFailedError(Exception):
    def __init__(self, operation: OperationRequest) -> None:  # noqa: D107
        op = operation.operation
        error = operation.error_code
        trace = operation.trace_id
        super().__init__(f"Operation {op} with trace {trace} failed: {error}")


class FailedToConnectError(Exception):
    pass


class NotConnectedError(Exception):
    pass
